program Example2;

uses
  Windows,
  Messages,
  dglOpenGL,
  DglShader;

const
  WND_TITLE = 'Image Space Lighting Demo';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms

  // User Constants
  FRUSTDIM: GLdouble = 100.0;
  FRUSTNEAR: GLdouble = 320.0;
  FRUSTFAR: GLdouble = 660.0;
  wall_mat: array [0..3] of GLfloat = (1.0, 1.0, 1.0, 1.0);
  lightpos: array [0..3] of GLfloat = (50.0, 50.0, -320.0, 1.0);
  sphere_mat: array [0..3] of GLfloat = (1.0, 0.5, 0.0, 1.0);
  cone_mat: array [0..3] of GLfloat = (0.0, 0.5, 1.0, 1.0);


var
  winwidth: integer = 512;           // window x size
  winheight: integer = 512;           // window y size
  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames

  // Textures
  Texture : GLuint;                  // Variable to hold the texture pointer

  // User vaiables
  xSpeed, ySpeed : glFloat;
  xAngle, yAngle : glFloat;

  // Mouse Variables
  MouseMove : boolean = false;        // Is the mouse moving?
  MoveXcoord, MoveYcoord : GLfloat;   // Current X and Y position of the cursor

  MouseButton : Integer = -1;         // mouse button down
  Xcoord, Ycoord : Integer;           // Current X and Y position where the mouse was clicked

  OpenGLInitialized : Boolean;

  LSphere: GLint = 1;
  LCone: GLint = 2;

  focus: GLFloat = 420.0;

  //shaders
  glslsimplevert: TGLSLShader;
  glslsimplefrag: TGLSLShader;
  glslsimpleprog: TGLSLProgram;

  //deferred rendering

  const
  light_cutoff = 0.01; //light intensity cutoff point
  NUM_LIGHTS = 400; //number of lights
  LIGHT_BRIGTNESS = 20; //brightness of lights

  var
  FBO_SAME_FORMAT: boolean = false;

  type
    TLight= record
      dist: double;
      pos: TglVector3f;
      col: TglVector3f;
    end;

  var
  lights : array[0..NUM_LIGHTS] of TLight;
  positionMap, normalMap, colorMap, attrMap, depthMap: GluInt;
  deferredFBO: GluInt;

  deferredvert: TGLSLShader;
  deferredfrag: TGLSLShader;
  deferredprog: TGLSLProgram;

  lightvert: TGLSLShader;
  lightfrag: TGLSLShader;
  lightprog: TGLSLProgram;

{$R *.RES}

{------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)  }
{------------------------------------------------------------------}
function IntToStr(Num : Integer) : String;  // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

procedure SolidCube(size: integer);
begin
  glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(size, 0, 0);
	glVertex3f(size, size, 0);
	glVertex3f(0, size, 0);
	glEnd();
end;

function emptyTexture(width: GluInt; height: GluInt; iformat: GluInt; format: GluInt; gltype: GluInt= GL_UNSIGNED_BYTE): GluInt;
var
	ret: GLuInt;
begin
	glGenTextures(1, @ret);
	glBindTexture(GL_TEXTURE_2D, ret);
	glTexImage2D(GL_TEXTURE_2D, 0, iformat, width, height, 0, format, gltype, nil);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	result := ret;
end;

procedure bindTexture(tex: gluint; glunit: gluint; target: gluint = GL_TEXTURE_2D);
begin
	glActiveTexture(GL_TEXTURE0 + glunit);
	glBindTexture(target, tex);
end;

procedure unbindTexture(glunit: gluint; target: gluint = GL_TEXTURE_2D);
begin
	glActiveTexture(GL_TEXTURE0 + glunit);
	glBindTexture(target, 0);
end;

procedure createDeferredMaps();
var
  status: GLenum;
  buffers: array[0..4] of GLEnum;
begin
	if FBO_SAME_FORMAT then
  begin
		positionMap := emptyTexture(winwidth, winheight, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		normalMap := emptyTexture(winwidth, winheight, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		colorMap := emptyTexture(winwidth, winheight, GL_RGBA_FLOAT32_ATI, GL_RGBA);
		attrMap := emptyTexture(winwidth, winheight, GL_RGBA_FLOAT32_ATI, GL_RGBA, GL_FLOAT);
		depthMap := emptyTexture(winwidth, winheight, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT);
	end else
  begin
		positionMap := emptyTexture(winwidth, winheight, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		normalMap := emptyTexture(winwidth, winheight, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		colorMap := emptyTexture(winwidth, winheight, GL_RGBA, GL_RGBA);
		attrMap := emptyTexture(winwidth, winheight, GL_RGB_FLOAT32_ATI, GL_RGB, GL_FLOAT);
		depthMap := emptyTexture(winwidth, winheight, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT);
	end;

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, deferredFBO);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, colorMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, positionMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, normalMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT3_EXT, GL_TEXTURE_2D, attrMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthMap, 0);


  buffers[0] := GL_COLOR_ATTACHMENT0_EXT;
  buffers[1] := GL_COLOR_ATTACHMENT1_EXT;
  buffers[2] := GL_COLOR_ATTACHMENT2_EXT;
  buffers[3] := GL_COLOR_ATTACHMENT3_EXT;
	glDrawBuffers(4, @buffers);

	glReadBuffer(GL_NONE);

	status := glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	if(not FBO_SAME_FORMAT) then
		if(status=GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT) then
    begin
			FBO_SAME_FORMAT := true;
			glDeleteTextures(1, @positionMap);
			glDeleteTextures(1, @normalMap);
			glDeleteTextures(1, @colorMap);
			glDeleteTextures(1, @attrMap);
			glDeleteTextures(1, @depthMap);

			createDeferredMaps();
			exit;
		end;

	//if(status <> GL_FRAMEBUFFER_COMPLETE_EXT){printf("ERROR: Your card does not support framebuffer objects with MRTs.\n"); exit(-1);}
end;

procedure genDeferred();
var i: integer;
    pos: TglVector3F;
begin

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, deferredFBO);
	glViewport(0, 0, winwidth, winheight);
	glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  deferredprog.Enable;

	deferredprog.setf('diff_coeff', 1);
	deferredprog.setf('phong_coeff', 0);
	deferredprog.setf('two_sided', 0.8);

	for i:=0 to NUM_LIGHTS-1 do
  begin
		glLoadIdentity();
		pos := lights[i].pos;
		glTranslated(pos[0], pos[1], pos[2]);
		glScaled(2, 2, 2.5);
		SolidCube(2);
		glLoadIdentity();
	end;

	deferredprog.setf('two_sided', 0);

	glBegin(GL_QUADS);
	glNormal3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(3000, 0, 0);
	glVertex3f(3000, 3000, 0);
	glVertex3f(0, 3000, 0);
	glEnd();

  deferredprog.Disable();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
end;

 procedure renderLight(alight: TLight);
 begin
 	lightprog.set3fv('lightPos', @alight.pos); //pass light pos to pixel shader
	lightprog.set3fv('lightCol', @alight.col); //pass light color (magnitude is power) to pixel shader
	lightprog.setf('dist', alight.dist); //pass the light's area of effect radius to pixel shader

  glLoadIdentity();
  glTranslated(alight.pos[0], alight.pos[1], alight.pos[2]);
  //glScaled(alight.dist, alight.dist, alight.dist);
  SolidCube(2);
  glLoadIdentity();
 end;

procedure renderLights();
var
  i: integer;
begin
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE); //set blend function
	glEnable(GL_CULL_FACE); //cull front faces
	glCullFace(GL_FRONT);
	glDisable(GL_DEPTH_TEST); //disable depth testing
	glDepthMask(false);

 	lightprog.Enable;

 	bindTexture(positionMap, 0); //bind position, normal and color textures from deferred shading pass
 	lightprog.seti('positionMap', 0);
 	bindTexture(normalMap, 1);
 	lightprog.seti('normalMap', 1);
 	bindTexture(colorMap, 2);
 	lightprog.seti('colorMap', 2);
 	bindTexture(attrMap, 3);
 	lightprog.seti('attrMap', 3);

	//camera.loadCameraPos(s, "camPos"); //set pixel shader attribute "camPos" to the camera pos (as it is needed for phong shading)
	//cur_shader = s;
  lightprog.set3f('camPos', 0, 0, 0);

	for i:=0 to NUM_LIGHTS-1 do
		renderLight(lights[i]);

	lightprog.Disable; //unbind and reset everything to desired values

	unbindTexture(0); //pos
	unbindTexture(1); //normal
	unbindTexture(2); //color
	unbindTexture(3); //attr

	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(true);
	glDisable(GL_BLEND);
end;

procedure glDrawScene();
begin


  //*
  //** Note: wall verticies are ordered so they are all front facing
  //** this lets me do back face culling to speed things up.
  //*/

  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @wall_mat);

  //* floor */
  //* make the floor textured */
  glEnable(GL_TEXTURE_2D);

  //*
  //** Since we want to turn texturing on for floor only, we have to
  //** make floor a separate glBegin()/glEnd() sequence. You can't
  //** turn texturing on and off between begin and end calls
  //*/
  glBegin(GL_QUADS);
  glNormal3f(0.0, 1.0, 0.0);
  glTexCoord2i(0, 0);
  glVertex3f(-100.0, -100.0, -320.0);
  glTexCoord2i(1, 0);
  glVertex3f( 100.0, -100.0, -320.0);
  glTexCoord2i(1, 1);
  glVertex3f( 100.0, -100.0, -640.0);
  glTexCoord2i(0, 1);
  glVertex3f(-100.0, -100.0, -640.0);
  glEnd();

  glDisable(GL_TEXTURE_2D);


(*
  //* walls */


  glBegin(GL_QUADS);

  //* left wall */
  glNormal3f(1.0, 0.0, 0.0);
  glVertex3f(-100.0, -100.0, -320.0);
  glVertex3f(-100.0, -100.0, -640.0);
  glVertex3f(-100.0,  100.0, -640.0);
  glVertex3f(-100.0,  100.0, -320.0);

  //* right wall */
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f( 100.0, -100.0, -320.0);
  glVertex3f( 100.0,  100.0, -320.0);
  glVertex3f( 100.0,  100.0, -640.0);
  glVertex3f( 100.0, -100.0, -640.0);

  //* ceiling */
  glNormal3f(0.0, -1.0, 0.0);
  glVertex3f(-100.0,  100.0, -320.0);
  glVertex3f(-100.0,  100.0, -640.0);
  glVertex3f( 100.0,  100.0, -640.0);
  glVertex3f( 100.0,  100.0, -320.0);

  //* back wall */
  glNormal3f(0.0, 0.0, 1.0);
  glVertex3f(-100.0, -100.0, -640.0);
  glVertex3f( 100.0, -100.0, -640.0);
  glVertex3f( 100.0,  100.0, -640.0);
  glVertex3f(-100.0,  100.0, -640.0);
  glEnd();
*)

  glslsimpleprog.Enable;
  glPushMatrix();
  glTranslatef(-80.0, -60.0, -420.0);
  glCallList(LSPHERE);
  glPopMatrix();
  glslsimpleprog.Disable;

  glPushMatrix();
  glTranslatef(-20.0, -80.0, -600.0);
  glCallList(LCONE);
  glPopMatrix();
end;

{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw();
var
  min, max, count,i , j  : integer;
  scale, dx, dy: GLfloat;
begin
  //Standard Render Mode
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-FRUSTDIM, FRUSTDIM, -FRUSTDIM, FRUSTDIM, FRUSTNEAR, FRUSTFAR);
  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
  glLoadIdentity();

  genDeferred();

  //glDrawScene();

  renderlights();

  glFlush();
end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit();
var
  sphere, cone, base : PGLUquadricObj;
  ln: integer;
  x: integer;
  y: integer;
begin

  xSpeed :=0.1;   // start with some movement
  ySpeed :=0.1;

	glClearDepth(1.0);
	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
	glDisable(GL_LIGHTING);
	glClearColor(0, 0, 0, 0);
	glColor4d(1, 1, 1, 1);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  //make display lists
  glNewList(LSPHERE, GL_COMPILE);
  //* make display lists for sphere and cone; for efficiency */
  sphere := gluNewQuadric();
  //glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @sphere_mat);
  gluSphere(sphere, 20.0, 20, 20);
  gluDeleteQuadric(sphere);
  glEndList();

  glNewList(LCONE, GL_COMPILE);
  cone := gluNewQuadric();
  base := gluNewQuadric();
  glRotatef(-90.0, 1.0, 0.0, 0.0);
  //glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @cone_mat);
  gluDisk(base, 0., 20., 20, 1);
  gluCylinder(cone, 20., 0., 60., 20, 20);
  gluDeleteQuadric(cone);
  gluDeleteQuadric(base);
  glEndList();

  //load simple glsl shader
  glslsimpleprog := TGLSLProgram.Create();
  glslsimplevert := TGLSLShader.Create('shaders\simple.vs');
  glslsimplefrag := TGLSLShader.Create('shaders\simple.fs', GL_FRAGMENT_SHADER_ARB);
  glslsimpleprog.Attach(glslsimplevert);
  glslsimpleprog.Attach(glslsimplefrag);
  glslsimpleprog.Link;

  //load deferred shader
  deferredprog := TGLSLProgram.Create();
  deferredvert := TGLSLShader.Create('shaders\deferred.vert');
  deferredfrag := TGLSLShader.Create('shaders\deferred.frag', GL_FRAGMENT_SHADER_ARB);
  deferredprog.Attach(glslsimplevert);
  deferredprog.Attach(glslsimplefrag);
  deferredprog.Link;

  //load light shader
  lightprog := TGLSLProgram.Create();
  lightvert := TGLSLShader.Create('shaders\light.vert');
  lightfrag := TGLSLShader.Create('shaders\light.frag', GL_FRAGMENT_SHADER_ARB);
  lightprog.Attach(glslsimplevert);
  lightprog.Attach(glslsimplefrag);
  lightprog.Link;

  //init deferred shading
  glGenFramebuffersEXT(1, @deferredFBO);
  createDeferredMaps;

  //init lights
  ln := 0;
  y:=-100;
	while y<3100 do //step 2900.0/(NUM_LIGHTS/2))
  begin
    x:=-20;
		while x <=20 do //x+= 20)
    begin
			if(ln < NUM_LIGHTS) then
      begin
        lights[ln].pos[0] := -x;
        lights[ln].pos[1] := -20;
        lights[ln].pos[2] := -y;//-420;

        lights[ln].col[0] := 1*LIGHT_BRIGTNESS;
        lights[ln].col[1] := 0.3*LIGHT_BRIGTNESS;
        lights[ln].col[2] := 0.086*LIGHT_BRIGTNESS;

        ln:=ln+1;
      end;
      x:=x+40;
    end;
    y:=y+round((2900/(NUM_LIGHTS/2)));
  end;

end;


{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
begin
  if (Height=0) then
     Height:=1;
  glViewport(0, 0, Width, Height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0,Width/Height,0.1,10000.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
end;

{------------------------------------------------------------------}
{  Processes all the mouse move                                  }
{------------------------------------------------------------------}
procedure ProcessMouseMove();
begin
  if MouseMove then
  begin
    // Add Code here to execute on mouse move
    MouseMove := False;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the mouse clicks                                  }
{------------------------------------------------------------------}
procedure ProcessMouseClick;
begin
  case MouseButton of
  1: // Left Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  2: // Right Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  3: // Middle Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0;  // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin
  if (keys[VK_UP])    then xspeed := xspeed - 0.002;
  if (keys[VK_DOWN])  then xspeed := xspeed + 0.002;
  if (keys[VK_RIGHT]) then yspeed := yspeed + 0.002;
  if (keys[VK_LEFT])  then yspeed := yspeed - 0.002;
end;


{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_MOUSEMOVE:
      begin
        MouseMove := True;
        MoveXcoord := LOWORD(lParam);
        MoveYcoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 1;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_RBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 2;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_MBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 3;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONUP,WM_RBUTTONUP,WM_MBUTTONUP:
      begin
        ReleaseCapture();   // above
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        if OpenGLInitialized then
         glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      begin

      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
      end;
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context
  if (not wglDeleteContext(h_RC)) then
  begin
    MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_RC := 0;
  end;

  // Attemps to release the device context
  if ((h_DC > 0) and (ReleaseDC(h_Wnd, h_DC) = 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;

  // Attempts to unregister the window class
  if (not UnRegisterClass('OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : DEVMODE;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
begin
  InitOpenGL;                               // New call to initialize and bind the OpenGL dll

  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE

  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := PixelDepth;           // Specifies the depth of the depth buffer
    cStencilBits    := PixelDepth;           // Specifies the depth of the stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;
  // Read And Assign Extentions
  ReadExtensions;
  ReadImplementationProperties;
  OpenGLInitialized := True;
  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(winwidth, winheight, FALSE, 32) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
  	TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime :=ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
      begin
        ProcessKeys;                      // Check for any other key Pressed
        ProcessMouseMove;                 // Check for mouse move
        ProcessMouseClick;                // Check for Mouse Click
      end;
    end;
  end;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.

